/* extension.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported init */

const { AccountsService, Clutter, Gio, GLib, GnomeBluetooth, GObject, St, Shell } = imports.gi;

const ExtensionUtils = imports.misc.extensionUtils;
const SystemActions = imports.misc.systemActions;
const BoxPointer = imports.ui.boxpointer;
const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;


class UserWidgetPopup extends PopupMenu.PopupMenu {
    constructor(userWidget) {
        super(userWidget, 0, St.Side.BOTTOM);
        this.actor.hide();
        this.actor.add_style_class_name('user-widget-popup');
        Main.uiGroup.add_actor(this.actor);
        
        this.userWidget = userWidget;

        this.blockSourceEvents = true;

        userWidget.connect('destroy', () => {
            this.destroy();
        });
        
        this._systemActions = new SystemActions.getDefault();
        this._buildMenu();
    }

    _buildMenu() {
        let bindFlags = GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE;
        let item;

        let app = this._settingsApp = Shell.AppSystem.get_default().lookup_app(
            'gnome-control-center.desktop'
        );
        if (app) {
            let [icon, name] = [app.app_info.get_icon().names[0],
                                app.get_name()];
            item = new PopupMenu.PopupImageMenuItem(name, icon);
            item.connect('activate', () => {
                this.itemActivated(BoxPointer.PopupAnimation.NONE);
                Main.overview.hide();
                this._settingsApp.activate();
            });
            this.addMenuItem(item);
        }

        item = new PopupMenu.PopupImageMenuItem(_('Lock'), 'changes-prevent-symbolic');
        item.connect('activate', () => {
            this.itemActivated(BoxPointer.PopupAnimation.NONE);
            this._systemActions.activateLockScreen();
        });
        this.addMenuItem(item);
        this._lockScreenItem = item;
        this._systemActions.bind_property(
            'can-lock-screen',
            this._lockScreenItem,
            'visible',
            bindFlags
        );

        this._sessionSubMenu = new PopupMenu.PopupSubMenuMenuItem(
            _('Power Off / Log Out'), true);
        this._sessionSubMenu.icon.icon_name = 'system-shutdown-symbolic';

        item = new PopupMenu.PopupMenuItem(_("Log Out"));
        item.connect('activate', () => {
            this.itemActivated(BoxPointer.PopupAnimation.NONE);
            this._systemActions.activateLogout();
        });
        this._sessionSubMenu.menu.addMenuItem(item);
        this._logoutItem = item;
        this._systemActions.bind_property('can-logout',
                                            this._logoutItem,
                                            'visible',
                                            bindFlags);

        item = new PopupMenu.PopupMenuItem(_("Switch User…"));
        item.connect('activate', () => {
            this.itemActivated(BoxPointer.PopupAnimation.NONE);
            this._systemActions.activateSwitchUser();
        });
        this._sessionSubMenu.menu.addMenuItem(item);
        this._loginScreenItem = item;
        this._systemActions.bind_property('can-switch-user',
                                            this._loginScreenItem,
                                            'visible',
                                            bindFlags);

        this._sessionSubMenu.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

        item = new PopupMenu.PopupMenuItem(_("Suspend"));
        item.connect('activate', () => {
            this.itemActivated(BoxPointer.PopupAnimation.NONE);
            this._systemActions.activateSuspend();
        });
        this._sessionSubMenu.menu.addMenuItem(item);
        this._suspendItem = item;
        this._systemActions.bind_property('can-suspend',
                                            this._suspendItem,
                                            'visible',
                                            bindFlags);

        item = new PopupMenu.PopupMenuItem(_("Power Off…"));
        item.connect('activate', () => {
            this.itemActivated(BoxPointer.PopupAnimation.NONE);
            this._systemActions.activatePowerOff();
        });
        this._sessionSubMenu.menu.addMenuItem(item);
        this._powerOffItem = item;
        this._systemActions.bind_property('can-power-off',
                                            this._powerOffItem,
                                            'visible',
                                            bindFlags);

        this.addMenuItem(this._sessionSubMenu);
    }

    destroy() {
        Main.uiGroup.remove_actor(this.actor);
        super.destroy();
    }
}

var OverviewUserWidget = GObject.registerClass(
class OverviewUserWidget extends St.Button {
    _init() {
        super._init({ 
            button_mask: St.ButtonMask.ONE,
            reactive: true,
            can_focus: true,
            style_class: 'overview-user-widget',
        });

        this.set_child(new St.Icon({
            icon_name: 'avatar-default-symbolic',
            height: 32,
        }));

        this._userManager = AccountsService.UserManager.get_default();
        this._user = this._userManager.get_user(GLib.get_user_name());
        this._userLoadedId = this._user.connect('notify::is-loaded', this._update.bind(this));
        this._userChangedId = this._user.connect('changed', this._update.bind(this));
        this._update();

        this.connect('clicked', this._onClick.bind(this));
        this._menu = null
        this._menuManager = new PopupMenu.PopupMenuManager(this);
    }

    _update() {
        let iconFile = this._user.get_icon_file();
        if (iconFile && !GLib.file_test(iconFile, GLib.FileTest.EXISTS))
            iconFile = null;

        if (iconFile) {
            let file = Gio.File.new_for_path(iconFile);
            this.child = null;
            this.style = 'background-image: url("%s"); background-size: cover;'.format(iconFile);
        } else {
            this.child = new St.Icon({ 
                icon_name: 'avatar-default-symbolic',
                height: 32
            });
        }
    }

    _onClick() {
        if (!this._menu) {
            this._menu = new UserWidgetPopup(this);
            
            let id = Main.overview.connect('hiding', () => {
                this._menu.close();
            });
            this.connect('destroy', () => {
                Main.overview.disconnect(id);
            });

            this._menuManager.addMenu(this._menu);
        }
        this._menu.open(false);

        return false;
    }

    destroy() {
        this._user.disconnect(this._userChangedId);
        this._user.disconnect(this._userLoadedId);
        super.destroy();
    }
}
);

var AppDisplayControlsBox = GObject.registerClass(
class AppDisplayControlsBox extends St.Widget {
    _init(appDisplay) {
        super._init({ layout_manager: new Clutter.BinLayout() });

        this.iconPadding = 24;
        
        this.appDisplay = appDisplay;
        this.appDisplayControlsParent = this.appDisplay._controls.get_parent();

        this.icon = new OverviewUserWidget();

        this.add_child(this.icon);

        this.appDisplay.remove_actor(this.appDisplayControlsParent);
        this.add_child(this.appDisplayControlsParent);
        this.appDisplay.add_actor(this);
    }

    destroy() {
        this.appDisplay.remove_actor(this);
        this.remove_actor(this.appDisplayControlsParent);
        this.appDisplay.add_actor(this.appDisplayControlsParent);

        this.icon.destroy();
        super.destroy();
    }

    vfunc_allocate(box, flags) {
        this.set_allocation(box, flags);

        let availWidth = box.x2 - box.x1;
        let availHeight = box.y2 - box.y1;

        let [, , iconNatWidth, iconNatHeight] = this.icon.get_preferred_size();

        let childBox = new Clutter.ActorBox();
        childBox.x1 = box.x1 + this.iconPadding;
        childBox.y1 = 0 ;
        childBox.x2 = box.x1 + iconNatWidth + this.iconPadding;
        childBox.y2 = iconNatHeight;
        this.icon.allocate(childBox, flags);

        let [, , controlsNatWidth, controlsNatHeight] = this.appDisplayControlsParent.get_preferred_size();
        childBox = new Clutter.ActorBox();
        childBox.x1 = (availWidth - controlsNatWidth)/2;
        childBox.y1 = 0;
        childBox.x2 = childBox.x1 + controlsNatWidth;
        childBox.y2 = childBox.y1 + controlsNatHeight;
        this.appDisplayControlsParent.allocate(childBox, flags);
    }
}
);


function _syncOverride() {
    // Do we have batteries or a UPS?
    let visible = this._proxy.IsPresent;
    if (visible) {
        this._item.show();
        this._percentageLabel.visible = this._desktopSettings.get_boolean(SHOW_BATTERY_PERCENTAGE);
    } else {
        // If there's no battery, then we use the power icon.
        this._item.hide();
        this._indicator.hide();
        this._percentageLabel.hide();
        return;
    }

    // The icons
    let chargingState = this._proxy.State == UPower.DeviceState.CHARGING
        ? '-charging' : '';
    let fillLevel = 10 * Math.floor(this._proxy.Percentage / 10);
    let icon;
    if (this._proxy.State == UPower.DeviceState.FULLY_CHARGED ||
        fillLevel === 100)
        icon = 'battery-level-100-charged-symbolic';
    else
        icon = 'battery-level-%d%s-symbolic'.format(fillLevel, chargingState);

    // Make sure we fall back to fallback-icon-name and not GThemedIcon's
    // default fallbacks
    let gicon = new Gio.ThemedIcon({
        name: icon,
        use_default_fallbacks: false,
    });

    this._indicator.gicon = gicon;
    this._item.icon.gicon = gicon;

    let fallbackIcon = this._proxy.IconName;
    this._indicator.fallback_icon_name = fallbackIcon;
    this._item.icon.fallback_icon_name = fallbackIcon;

    // The icon label
    let label;
    if (this._proxy.State == UPower.DeviceState.FULLY_CHARGED)
        label = _("%d\u2009%%").format(100);
    else
        label = _("%d\u2009%%").format(this._proxy.Percentage);
    this._percentageLabel.clutter_text.set_markup('<span size="smaller">' + label + '</span>');

    // The status label
    this._item.label.text = this._getStatus();
}

function _updateSessionSubMenuOverride() {
    if (Main.sessionMode.hasOverview) {
        this._sessionSubMenu.visible = false;
    } else {
        this._sessionSubMenu.visible =
            this._loginScreenItem.visible ||
            this._logoutItem.visible ||
            this._suspendItem.visible ||
            this._powerOffItem.visible;
    }
}

function _sessionUpdatedOverride() {
    if (Main.sessionMode.hasOverview) {
        this._sessionSubMenu.visible = false;
    } else {
        this._settingsItem.visible = Main.sessionMode.allowSettings;
    }

    this._lockScreenItem.visible = this._settingsItem.visible && this._systemActions.can_lock_screen;
}

class Extension {
    constructor() {
        this.layout = null;
        this._aggregateMenu = Main.panel.statusArea['aggregateMenu'];
        this._systemIndicator = this._aggregateMenu._system;
        this._systemIndicatorUpdateSessionSubMenuFunction = this._systemIndicator._updateSessionSubMenu;
        this._systemIndicatorSessionUpdatedFunction = this._systemIndicator._sessionUpdated;
        this._powerIndicator = this._aggregateMenu._power;
        this._powerIndicatorSyncFunction = this._powerIndicator._sync;
        this._arrowIcon = this._aggregateMenu._indicators.get_last_child();
    }

    enable() {
        let overviewControls = Main.overview._overview._controls;
        let viewSelector = overviewControls.viewSelector;
        let appDisplay = viewSelector.appDisplay;
        this.layout = new AppDisplayControlsBox(appDisplay);

        this._aggregateMenu._indicators.remove_actor(this._arrowIcon);

        this._powerIndicator._sync = _syncOverride.bind(this._powerIndicator);
        this._powerIndicator._sync();
        
        this._systemIndicator._orientationLockItem.hide();
        this._systemIndicator._settingsItem.hide();
        this._systemIndicator._lockScreenItem.hide();
        this._systemIndicator._sessionSubMenu.hide();
        this._systemIndicator._updateSessionSubMenu = _updateSessionSubMenuOverride.bind(this._systemIndicator);
        this._systemIndicator._sessionUpdated = _sessionUpdatedOverride.bind(this._systemIndicator);
    }

    disable() {
        this.layout.destroy();

        this._systemIndicator._sessionUpdated = this._systemIndicatorSessionUpdatedFunction;
        this._systemIndicator._updateSessionSubMenu = this._systemIndicatorUpdateSessionSubMenuFunction;
        this._systemIndicator._orientationLockItem.visible = this._systemIndicator._systemActions.can_lock_orientation;
        this._systemIndicator._sessionUpdated();
        this._systemIndicator._lockScreenItem.visible = this._systemIndicator._systemActions.can_lock_screen;
        this._systemIndicator._sessionSubMenu.show();

        this._powerIndicator._sync = this._powerIndicatorSyncFunction;
        this._powerIndicator._indicator.show();
        this._powerIndicator._sync();

        this._aggregateMenu._indicators.add_child(this._arrowIcon);
    }
}

function init() {
    return new Extension();
}
